@@ -21,17 +21,15 @@ import (
 	"fmt"
 	"io"
 	"os"
-	"os/exec"
-	"path/filepath"
-	"sort"
 	"strings"
 	"time"
 
+	"cuelang.org/go/cue"
 	"cuelang.org/go/cue/cuecontext"
 	"github.com/fluxcd/pkg/ssa"
+	"github.com/go-logr/logr"
 	"github.com/spf13/cobra"
 	"k8s.io/apimachinery/pkg/apis/meta/v1/unstructured"
-	"sigs.k8s.io/yaml"
 
 	apiv1 "github.com/stefanprodan/timoni/api/v1alpha1"
 	"github.com/stefanprodan/timoni/internal/engine"
@@ -63,13 +61,14 @@ var bundleApplyCmd = &cobra.Command{
 }
 
 type bundleApplyFlags struct {
-	pkg    flags.Package
-	files  []string
-	dryrun bool
-	diff   bool
-	wait   bool
-	force  bool
-	creds  flags.Credentials
+	pkg                flags.Package
+	files              []string
+	dryrun             bool
+	diff               bool
+	wait               bool
+	force              bool
+	overwriteOwnership bool
+	creds              flags.Credentials
 }
 
 var bundleApplyArgs bundleApplyFlags
@@ -80,6 +79,8 @@ func init() {
 		"The local path to bundle.cue files.")
 	bundleApplyCmd.Flags().BoolVar(&bundleApplyArgs.force, "force", false,
 		"Recreate immutable Kubernetes resources.")
+	bundleApplyCmd.Flags().BoolVar(&bundleApplyArgs.overwriteOwnership, "overwrite-ownership", false,
+		"Overwrite instance ownership, if any instances are owned by other Bundles.")
 	bundleApplyCmd.Flags().BoolVar(&bundleApplyArgs.dryrun, "dry-run", false,
 		"Perform a server-side apply dry run.")
 	bundleApplyCmd.Flags().BoolVar(&bundleApplyArgs.diff, "diff", false,
@@ -90,17 +91,8 @@ func init() {
 	bundleCmd.AddCommand(bundleApplyCmd)
 }
 
-func runBundleApplyCmd(cmd *cobra.Command, args []string) error {
-	bundleSchema, err := os.CreateTemp("", "schema.*.cue")
-	if err != nil {
-		return err
-	}
-	defer os.Remove(bundleSchema.Name())
-	if _, err := bundleSchema.WriteString(apiv1.BundleSchema); err != nil {
-		return err
-	}
-
-	files := append(bundleApplyArgs.files, bundleSchema.Name())
+func runBundleApplyCmd(cmd *cobra.Command, _ []string) error {
+	files := bundleApplyArgs.files
 	for i, file := range files {
 		if file == "-" {
 			path, err := saveReaderToFile(cmd.InOrStdin())
@@ -114,30 +106,60 @@ func runBundleApplyCmd(cmd *cobra.Command, args []string) error {
 		}
 	}
 
-	ctx := cuecontext.New()
-	bm := engine.NewBundleBuilder(ctx, files)
+	tmpDir, err := os.MkdirTemp("", apiv1.FieldManager)
+	if err != nil {
+		return err
+	}
+	defer os.RemoveAll(tmpDir)
+
+	cuectx := cuecontext.New()
+	bm := engine.NewBundleBuilder(cuectx, files)
+
+	if err := bm.InitWorkspace(tmpDir); err != nil {
+		return err
+	}
 
 	v, err := bm.Build()
 	if err != nil {
-		return err
+		return describeErr(tmpDir, "failed to build bundle", err)
 	}
 
-	instances, err := bm.GetInstances(v)
+	bundle, err := bm.GetBundle(v)
 	if err != nil {
 		return err
 	}
 
-	for _, instance := range instances {
-		logger.Printf("applying instance %s", instance.Name)
-		if err := applyBundleInstance(instance); err != nil {
+	log := LoggerFrom(cmd.Context(), "bundle", bundle.Name)
+
+	if !bundleApplyArgs.overwriteOwnership {
+		err = bundleInstancesOwnershipConflicts(bundle.Instances)
+		if err != nil {
 			return err
 		}
 	}
 
+	log.Info(fmt.Sprintf("applying %v instance(s)", len(bundle.Instances)))
+
+	ctx, cancel := context.WithTimeout(cmd.Context(), rootArgs.timeout)
+	defer cancel()
+
+	for _, instance := range bundle.Instances {
+		log.Info(fmt.Sprintf("applying instance %s", instance.Name))
+		if err := applyBundleInstance(logr.NewContext(ctx, log), cuectx, instance); err != nil {
+			return err
+		}
+	}
+
+	if bundleApplyArgs.dryrun || bundleApplyArgs.diff {
+		log.Info(fmt.Sprintf("applied %v instance(s) (server dry run)", len(bundle.Instances)))
+	} else {
+		log.Info("applied successfully")
+	}
+
 	return nil
 }
 
-func applyBundleInstance(instance engine.BundleInstance) error {
+func applyBundleInstance(ctx context.Context, cuectx *cue.Context, instance engine.BundleInstance) error {
 	moduleVersion := instance.Module.Version
 	sourceURL := fmt.Sprintf("%s:%s", instance.Module.Repository, instance.Module.Version)
 
@@ -146,7 +168,8 @@ func applyBundleInstance(instance engine.BundleInstance) error {
 		moduleVersion = "@" + instance.Module.Digest
 	}
 
-	logger.Printf("pulling %s", sourceURL)
+	log := LoggerFrom(ctx, "instance", instance.Name)
+	log.Info(fmt.Sprintf("pulling %s", sourceURL))
 
 	tmpDir, err := os.MkdirTemp("", apiv1.FieldManager)
 	if err != nil {
@@ -174,7 +197,6 @@ func applyBundleInstance(instance engine.BundleInstance) error {
 			mod.Digest, instance.Module.Version, instance.Module.Digest)
 	}
 
-	cuectx := cuecontext.New()
 	builder := engine.NewModuleBuilder(
 		cuectx,
 		instance.Name,
@@ -183,40 +205,35 @@ func applyBundleInstance(instance engine.BundleInstance) error {
 		bundleApplyArgs.pkg.String(),
 	)
 
+	if err := builder.WriteSchemaFile(); err != nil {
+		return err
+	}
+
 	mod.Name, err = builder.GetModuleName()
 	if err != nil {
 		return err
 	}
 
-	logger.Printf("using module %s version %s", mod.Name, mod.Version)
+	log.Info(fmt.Sprintf("using module %s version %s", mod.Name, mod.Version))
 
-	err = builder.WriteValuesFile(instance.Values)
+	err = builder.WriteValuesFileWithDefaults(instance.Values)
 	if err != nil {
 		return err
 	}
 
 	buildResult, err := builder.Build()
 	if err != nil {
-		return fmt.Errorf("failed to build instance, error: %w", err)
-	}
-
-	apiVer, err := builder.GetAPIVersion(buildResult)
-	if err != nil {
-		return err
-	}
-
-	if apiVer != apiv1.GroupVersion.Version {
-		return fmt.Errorf("API version %s not supported, must be %s", apiVer, apiv1.GroupVersion.Version)
+		return describeErr(fetcher.GetModuleRoot(), "failed to build instance", err)
 	}
 
 	finalValues, err := builder.GetValues(buildResult)
 	if err != nil {
-		return fmt.Errorf("failed to extract values, error: %w", err)
+		return fmt.Errorf("failed to extract values: %w", err)
 	}
 
 	bundleApplySets, err := builder.GetApplySets(buildResult)
 	if err != nil {
-		return fmt.Errorf("failed to extract objects, error: %w", err)
+		return fmt.Errorf("failed to extract objects: %w", err)
 	}
 
 	var objects []*unstructured.Unstructured
@@ -231,110 +248,91 @@ func applyBundleInstance(instance engine.BundleInstance) error {
 
 	rm.SetOwnerLabels(objects, instance.Name, instance.Namespace)
 
-	ctx, cancel := context.WithTimeout(context.Background(), rootArgs.timeout)
-	defer cancel()
-
 	exists := false
 	sm := runtime.NewStorageManager(rm)
-	if _, err := sm.Get(ctx, instance.Name, instance.Namespace); err == nil {
+	if _, err = sm.Get(ctx, instance.Name, instance.Namespace); err == nil {
 		exists = true
 	}
 
+	nsExists, err := sm.NamespaceExists(ctx, instance.Namespace)
+	if err != nil {
+		return fmt.Errorf("instance init failed: %w", err)
+	}
+
 	if bundleApplyArgs.dryrun || bundleApplyArgs.diff {
-		diffOpts := ssa.DefaultDiffOptions()
-		sort.Sort(ssa.SortableUnstructureds(objects))
-		for _, r := range objects {
-			change, liveObject, mergedObject, err := rm.Diff(ctx, r, diffOpts)
-			if err != nil {
-				logger.Println(err)
-				continue
-			}
-
-			logger.Println(change.String(), "(server dry run)")
-
-			if bundleApplyArgs.diff && change.Action == ssa.ConfiguredAction {
-				liveYAML, _ := yaml.Marshal(liveObject)
-				liveFile := filepath.Join(tmpDir, "live.yaml")
-				if err := os.WriteFile(liveFile, liveYAML, 0644); err != nil {
-					return err
-				}
-
-				mergedYAML, _ := yaml.Marshal(mergedObject)
-				mergedFile := filepath.Join(tmpDir, "merged.yaml")
-				if err := os.WriteFile(mergedFile, mergedYAML, 0644); err != nil {
-					return err
-				}
-
-				out, _ := exec.Command("diff", "-N", "-u", liveFile, mergedFile).Output()
-				for i, line := range strings.Split(string(out), "\n") {
-					if i > 1 && len(line) > 0 {
-						logger.Println(line)
-					}
-				}
-			}
+		if !nsExists {
+			log.Info(fmt.Sprintf("Namespace/%s created (server dry run)", instance.Namespace))
+		}
+		if err := instanceDryRun(logr.NewContext(ctx, log), rm, objects, nsExists, tmpDir, bundleApplyArgs.diff); err != nil {
+			return err
 		}
 
-		logger.Println("bundle applied successfully")
+		log.Info("applied successfully (server dry run)")
 		return nil
 	}
 
 	im := runtime.NewInstanceManager(instance.Name, instance.Namespace, finalValues, *mod)
 
+	if im.Instance.Labels == nil {
+		im.Instance.Labels = make(map[string]string)
+	}
+	im.Instance.Labels[apiv1.BundleNameLabelKey] = instance.Bundle
+
 	if err := im.AddObjects(objects); err != nil {
-		return fmt.Errorf("adding objects to instance failed, error: %w", err)
+		return fmt.Errorf("adding objects to instance failed: %w", err)
 	}
 
 	if !exists {
-		logger.Printf("installing %s in namespace %s", instance.Name, instance.Namespace)
-
-		nsExists, err := sm.NamespaceExists(ctx, instance.Namespace)
-		if err != nil {
-			return fmt.Errorf("instance init failed, error: %w", err)
-		}
+		log.Info(fmt.Sprintf("installing %s in namespace %s", instance.Name, instance.Namespace))
 
 		if err := sm.Apply(ctx, &im.Instance, true); err != nil {
-			return fmt.Errorf("instance init failed, error: %w", err)
+			return fmt.Errorf("instance init failed: %w", err)
 		}
 
 		if !nsExists {
-			logger.Printf("Namespace/%s created", instance.Namespace)
+			log.Info(fmt.Sprintf("Namespace/%s created", instance.Namespace))
 		}
 	} else {
-		logger.Printf("upgrading %s in namespace %s", instance.Name, instance.Namespace)
+		log.Info(fmt.Sprintf("upgrading %s in namespace %s", instance.Name, instance.Namespace))
 	}
 
-	bundleApplyOpts := runtime.ApplyOptions(bundleApplyArgs.force, time.Minute)
+	applyOpts := runtime.ApplyOptions(bundleApplyArgs.force, rootArgs.timeout)
+	waitOptions := ssa.WaitOptions{
+		Interval: 5 * time.Second,
+		Timeout:  rootArgs.timeout,
+	}
 
 	for _, set := range bundleApplySets {
 		if len(bundleApplySets) > 1 {
-			logger.Println("applying", set.Name)
+			log.Info(fmt.Sprintf("applying %s", set.Name))
 		}
 
-		cs, err := rm.ApplyAllStaged(ctx, set.Objects, bundleApplyOpts)
+		cs, err := rm.ApplyAllStaged(ctx, set.Objects, applyOpts)
 		if err != nil {
 			return err
 		}
 		for _, change := range cs.Entries {
-			logger.Println(change.String())
+			log.Info(change.String())
 		}
 
 		if bundleApplyArgs.wait {
-			logger.Println(fmt.Sprintf("waiting for %v resource(s) to become ready...", len(set.Objects)))
-			err = rm.Wait(set.Objects, ssa.DefaultWaitOptions())
+			spin := StartSpinner(fmt.Sprintf("waiting for %v resource(s) to become ready...", len(set.Objects)))
+			err = rm.Wait(set.Objects, waitOptions)
+			spin.Stop()
 			if err != nil {
 				return err
 			}
-			logger.Println("resources are ready")
+			log.Info("resources are ready")
 		}
 	}
 
 	staleObjects, err := sm.GetStaleObjects(ctx, &im.Instance)
 	if err != nil {
-		return fmt.Errorf("getting stale objects failed, error: %w", err)
+		return fmt.Errorf("getting stale objects failed: %w", err)
 	}
 
 	if err := sm.Apply(ctx, &im.Instance, true); err != nil {
-		return fmt.Errorf("storing instance failed, error: %w", err)
+		return fmt.Errorf("storing instance failed: %w", err)
 	}
 
 	var deletedObjects []*unstructured.Unstructured
@@ -342,29 +340,58 @@ func applyBundleInstance(instance engine.BundleInstance) error {
 		deleteOpts := runtime.DeleteOptions(instance.Name, instance.Namespace)
 		changeSet, err := rm.DeleteAll(ctx, staleObjects, deleteOpts)
 		if err != nil {
-			return fmt.Errorf("prunning objects failed, error: %w", err)
+			return fmt.Errorf("prunning objects failed: %w", err)
 		}
 		deletedObjects = runtime.SelectObjectsFromSet(changeSet, ssa.DeletedAction)
 		for _, change := range changeSet.Entries {
-			logger.Println(change.String())
+			log.Info(change.String())
 		}
 	}
 
 	if bundleApplyArgs.wait {
 		if len(deletedObjects) > 0 {
-			logger.Printf("waiting for %v resource(s) to be finalized...", len(deletedObjects))
-			err = rm.WaitForTermination(deletedObjects, ssa.DefaultWaitOptions())
+			spin := StartSpinner(fmt.Sprintf("waiting for %v resource(s) to be finalized...", len(deletedObjects)))
+			err = rm.WaitForTermination(deletedObjects, waitOptions)
+			spin.Stop()
 			if err != nil {
-				return fmt.Errorf("wating for termination failed, error: %w", err)
+				return fmt.Errorf("wating for termination failed: %w", err)
 			}
 
-			logger.Println("all resources are ready")
+			log.Info("all resources are ready")
 		}
 	}
 
 	return nil
 }
 
+func bundleInstancesOwnershipConflicts(bundleInstances []engine.BundleInstance) error {
+	var conflicts []string
+	rm, err := runtime.NewResourceManager(kubeconfigArgs)
+	if err != nil {
+		return err
+	}
+
+	ctx, cancel := context.WithTimeout(context.Background(), rootArgs.timeout)
+	defer cancel()
+
+	sm := runtime.NewStorageManager(rm)
+	for _, instance := range bundleInstances {
+		if existingInstance, err := sm.Get(ctx, instance.Name, instance.Namespace); err == nil {
+			currentOwnerBundle := existingInstance.Labels[apiv1.BundleNameLabelKey]
+			if currentOwnerBundle == "" {
+				conflicts = append(conflicts, fmt.Sprintf("instance \"%s\" exists and is managed by no bundle", instance.Name))
+			} else if currentOwnerBundle != instance.Bundle {
+				conflicts = append(conflicts, fmt.Sprintf("instance \"%s\" exists and is managed by another bundle \"%s\"", instance.Name, currentOwnerBundle))
+			}
+		}
+	}
+	if len(conflicts) > 0 {
+		return fmt.Errorf("instance ownership conflicts encountered. Apply with \"--overwrite-ownership\" to gain instance ownership. Conflicts: %s", strings.Join(conflicts, "; "))
+	}
+
+	return nil
+}
+
 func saveReaderToFile(reader io.Reader) (string, error) {
 	f, err := os.CreateTemp("", "*.cue")
 	if err != nil {
