@@ -41,6 +41,9 @@ deletes all their Kubernetes resources from the cluster.'.
 	Example: `  # Uninstall all instances in a bundle
   timoni bundle delete -f bundle.cue
 
+  # Uninstall all instances in a named bundle
+  timoni bundle delete --name podinfo
+
   # Uninstall all instances without waiting for finalisation
   timoni bundle delete -f bundle.cue --wait=false
 
@@ -51,9 +54,11 @@ deletes all their Kubernetes resources from the cluster.'.
 }
 
 type bundleDelFlags struct {
-	files  []string
-	wait   bool
-	dryrun bool
+	files         []string
+	allNamespaces bool
+	wait          bool
+	dryrun        bool
+	name          string
 }
 
 var bundleDelArgs bundleDelFlags
@@ -65,10 +70,53 @@ func init() {
 		"Wait for the deleted Kubernetes objects to be finalized.")
 	bundleDelCmd.Flags().BoolVar(&bundleDelArgs.dryrun, "dry-run", false,
 		"Perform a server-side delete dry run.")
+	bundleDelCmd.Flags().StringVar(&bundleDelArgs.name, "name", "",
+		"Name of the bundle to delete.")
+	bundleDelCmd.Flags().BoolVarP(&bundleDelArgs.allNamespaces, "all-namespaces", "A", false,
+		"Delete the requested Bundle across all namespaces.")
 	bundleCmd.AddCommand(bundleDelCmd)
 }
 
-func runBundleDelCmd(cmd *cobra.Command, args []string) error {
+func runBundleDelCmd(cmd *cobra.Command, _ []string) error {
+	ctx, cancel := context.WithTimeout(cmd.Context(), rootArgs.timeout)
+	defer cancel()
+
+	if bundleDelArgs.name != "" {
+		return deleteBundleByName(ctx, bundleDelArgs.name)
+	}
+
+	return deleteBundleFromFile(ctx, cmd)
+}
+
+func deleteBundleByName(ctx context.Context, bundle string) error {
+	sm, err := runtime.NewResourceManager(kubeconfigArgs)
+	if err != nil {
+		return err
+	}
+
+	log := LoggerFrom(ctx, "bundle", bundle)
+	iStorage := runtime.NewStorageManager(sm)
+
+	instances, err := iStorage.List(ctx, "", bundle)
+	if err != nil {
+		return err
+	}
+
+	for _, instance := range instances {
+		log.Info(fmt.Sprintf("deleting instance %s from bundle %s", instance.Name, bundleDelArgs.name))
+		if err := deleteBundleInstance(ctx, engine.BundleInstance{
+			Bundle:    bundle,
+			Name:      instance.Name,
+			Namespace: instance.Namespace,
+		}, bundleDelArgs.wait, bundleDelArgs.dryrun); err != nil {
+			return err
+		}
+	}
+
+	return nil
+}
+
+func deleteBundleFromFile(ctx context.Context, cmd *cobra.Command) error {
 	bundleSchema, err := os.CreateTemp("", "schema.*.cue")
 	if err != nil {
 		return err
@@ -92,26 +140,28 @@ func runBundleDelCmd(cmd *cobra.Command, args []string) error {
 		}
 	}
 
-	ctx := cuecontext.New()
-	bm := engine.NewBundleBuilder(ctx, files)
+	cuectx := cuecontext.New()
+	bm := engine.NewBundleBuilder(cuectx, files)
 
 	v, err := bm.Build()
 	if err != nil {
 		return err
 	}
 
-	instances, err := bm.GetInstances(v)
+	bundle, err := bm.GetBundle(v)
 	if err != nil {
 		return err
 	}
 
-	if len(instances) == 0 {
+	log := LoggerFrom(ctx, "bundle", bundle.Name)
+
+	if len(bundle.Instances) == 0 {
 		return fmt.Errorf("no instances found in bundle")
 	}
 
-	for _, instance := range instances {
-		logger.Printf("deleting instance %s", instance.Name)
-		if err := deleteBundleInstance(instance, bundleDelArgs.wait, bundleDelArgs.dryrun); err != nil {
+	for _, instance := range bundle.Instances {
+		log.Info(fmt.Sprintf("deleting instance %s", instance.Name))
+		if err := deleteBundleInstance(ctx, instance, bundleDelArgs.wait, bundleDelArgs.dryrun); err != nil {
 			return err
 		}
 	}
@@ -119,7 +169,9 @@ func runBundleDelCmd(cmd *cobra.Command, args []string) error {
 	return nil
 }
 
-func deleteBundleInstance(instance engine.BundleInstance, wait bool, dryrun bool) error {
+func deleteBundleInstance(ctx context.Context, instance engine.BundleInstance, wait bool, dryrun bool) error {
+	log := LoggerFrom(ctx, "bundle", instance.Bundle)
+
 	sm, err := runtime.NewResourceManager(kubeconfigArgs)
 	if err != nil {
 		return err
@@ -144,26 +196,26 @@ func deleteBundleInstance(instance engine.BundleInstance, wait bool, dryrun bool
 
 	if dryrun {
 		for _, object := range objects {
-			logger.Println(fmt.Sprintf(
+			log.Info(fmt.Sprintf(
 				"%s/%s/%s deleted (dry run)",
 				object.GetKind(), object.GetNamespace(), object.GetName()))
 		}
 		return nil
 	}
 
-	logger.Println(fmt.Sprintf("deleting %v resource(s)...", len(objects)))
+	log.Info(fmt.Sprintf("deleting %v resource(s)...", len(objects)))
 	hasErrors := false
 	cs := ssa.NewChangeSet()
 	for _, object := range objects {
 		deleteOpts := runtime.DeleteOptions(instance.Name, instance.Namespace)
 		change, err := sm.Delete(ctx, object, deleteOpts)
 		if err != nil {
-			logger.Println(`✗`, err)
+			log.Error(err, "deletion failed")
 			hasErrors = true
 			continue
 		}
 		cs.Add(*change)
-		logger.Println(change.String())
+		log.Info(fmt.Sprintf(change.String()))
 	}
 
 	if hasErrors {
@@ -178,12 +230,13 @@ func deleteBundleInstance(instance engine.BundleInstance, wait bool, dryrun bool
 	if wait && len(deletedObjects) > 0 {
 		waitOpts := ssa.DefaultWaitOptions()
 		waitOpts.Timeout = rootArgs.timeout
-		logger.Printf("waiting for %v resource(s) to be finalized...", len(deletedObjects))
+		spin := StartSpinner(fmt.Sprintf("waiting for %v resource(s) to be finalized...", len(deletedObjects)))
 		err = sm.WaitForTermination(deletedObjects, waitOpts)
+		spin.Stop()
 		if err != nil {
 			return err
 		}
-		logger.Println("all resources have been deleted")
+		log.Info("all resources have been deleted")
 	}
 
 	return nil
